#include "ystring.h"
#include <process.h>
#include <io.h>
#include <sys/locking.h>

int file_isregular(int fd);
static int LockingX(int fd,int type,int bytes);
int lock_dbg = 0;
int lock_ext = 1;

static int Locking(int fd,int type,int bytes)
{ 	int off,rcode;

	if( lock_ext ){
		return LockingX(fd,type,bytes);
	}

	rcode = -1;
	if( 0 <= (off = tell(fd)) )
	if( 0 <= lseek(fd,0,0) ){
		rcode = locking(fd,type,bytes);
		lseek(fd,off,0);
	}
	if( rcode == -1 && !file_isregular(fd) ){
		/* treat as success, 
		 * mainly for lock_TIMEOUT not to wait non-lockable file ... ?
		 */
		rcode = 0;
	}
	return rcode;
}

int SHlockB(int fd){	return	Locking(fd,_LK_RLCK,1); }
int SHlockNB(int fd){	return	Locking(fd,_LK_NBRLCK,1); }
int EXlockB(int fd){	return	Locking(fd,_LK_LOCK,1); }
int EXlockNB(int fd){	return	Locking(fd,_LK_NBLCK,1); }
int UNlock(int fd){	return	Locking(fd,_LK_UNLCK,1); }


int porting_dbg(const char *fmt,...);
static void lklog(int fd,int type,int bytes,const char *fmt,...){
	VARGS(16,fmt);
	CStr(msg,1024);
	const char *st;
	int woff;

	if( !file_isregular(fd) ) /* not the tty for LOGFILE */
		return;

	woff = tell(fd);
	switch(type){
		case _LK_RLCK:  st = "RLCK";   break;
		case _LK_NBRLCK:st = "NBRLCK"; break;
		case _LK_LOCK:  st = "LOCK";   break;
		case _LK_NBLCK: st = "NBLCK";  break;
		case _LK_UNLCK: st = "UNLCK";  break;
	}
	sprintf(msg,"Locking(%d,%d %s,%d) ",fd,type,st,bytes);
	Xsprintf(TVStr(msg),fmt,VA8);
	porting_dbg("%s",msg);
}

static int LockingX(int fd,int type,int xbytes){
	int shared;
 	int off,rcode;
	int serrno;
	int pid;
	int bytes = 0;
	int poff;

	shared = (type == _LK_RLCK || type == _LK_NBRLCK);
	rcode = -1;
	if( 0 <= (off = tell(fd)) ){
		pid = _getpid() & 0xFFFF;
		if( shared ){
			poff = pid;
			bytes = 1;
		}else{
			poff = 0;
			bytes = 0x10000;
		}
		if( 0 <= lseek(fd,poff,0) ){
			rcode = locking(fd,type,bytes);
			if( rcode != 0 && type == _LK_UNLCK ){
				if( shared ){
					poff = 0;
					bytes = 0x10000;
				}else{
					poff = pid;
					bytes = 1;
				}
				lseek(fd,poff,0);
				errno = 0;
				rcode = locking(fd,type,bytes);
			}
			serrno = errno;
			if( type != _LK_NBRLCK && type != _LK_NBLCK )
			if( rcode != 0 || lock_dbg ){
				lklog(fd,type,bytes,"=%d err=%d",rcode,errno);
			}
			lseek(fd,off,0);
			errno = serrno;
		}else{
			lklog(fd,type,bytes,"seek error");
			errno = 0;
		}
	}else{
		lklog(fd,type,bytes,"tell error");
		errno = 0;
	}
	if( rcode == -1 && !file_isregular(fd) ){
		/* treat as success, 
		 * mainly for lock_TIMEOUT not to wait non-lockable file ... ?
		 */
		rcode = 0;
	}
	return rcode;
}
