/*////////////////////////////////////////////////////////////////////////
Copyright (c) 1996 Electrotechnical Laboratry (ETL), AIST, MITI

Permission to use, copy, modify, and distribute this material for any
purpose and without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies, and
that the name of ETL not be used in advertising or publicity pertaining
to this material without the specific, prior written permission of an
authorized representative of ETL.
ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	fstat.c
Author:		Yutaka Sato <ysato@etl.go.jp>
Description:
History:
	961010	extracted from misc.c
//////////////////////////////////////////////////////////////////////#*/
#include <stdio.h>
#include "ystring.h"
#include "file.h"
#include <stdlib.h>
#include <time.h>
int INHERENT_lstat();

#include <sys/types.h>
#include <sys/stat.h>
#ifdef S_ISUID /* set UID on execution */
#include <pwd.h>
#endif


#if defined(__CYGWIN__)
#include <unistd.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>

#ifndef S_ISLNK
#ifndef S_IFLNK
#define S_ISLNK(m)	0
#else
#define S_ISLNK(m)  (((m)&S_IFMT) == S_IFLNK)
#endif
#endif

#ifndef S_ISDIR
#define S_ISDIR(m)  (((m)&S_IFMT) == S_IFDIR)
#endif
#ifndef S_ISREG
#define S_ISREG(m)  (((m)&S_IFMT) == S_IFREG)
#endif
#ifndef S_ISFIFO
#ifdef S_IFIFO
#define S_ISFIFO(m) (((m)&S_IFMT) == S_IFIFO)
#else
#define S_ISFIFO(m) 0
#endif
#endif

#if defined(S_ISUID) && defined(S_IRWXU)
#define GETEUID()	geteuid()
#define GETEGID()	getegid()
#define STUID(st)	st.st_uid
#define STGID(st)	st.st_gid
#else
#define GETEUID()	0
#define GETEGID()	0
#define STUID(st)	0
#define STGID(st)	0
#define S_IRWXU		(S_IREAD | S_IWRITE | S_IEXEC)
#define S_IRWXG		(S_IREAD | S_IWRITE | S_IEXEC)
#define S_IRWXO		(S_IREAD | S_IWRITE | S_IEXEC)
#endif

/*
 * This check should be done by access() function, but the function
 * can be platform dependent ??
 */
int access_RWX(PCStr(path))
{	struct stat st;
	int mode,uid,gid;

#if defined(__CYGWIN__)
	if( access(path,R_OK|W_OK|X_OK) == 0 )
		return 0;
#endif

	if( stat(path,&st) == 0 ){
		mode = st.st_mode;
		if( STUID(st) == GETEUID() && (mode & S_IRWXU) == S_IRWXU )
			return 0;
		if( STGID(st) == GETEGID() && (mode & S_IRWXG) == S_IRWXG )
			return 0;
		if( (mode & S_IRWXO) == S_IRWXO )
			return 0;
	}
	return -1;
}

int valid_fdl(PVStr(vfd))
{	int fd,nvfd,lastv;
	refQStr(vp,vfd); /**/

	nvfd = 0;
	lastv = 0;
	for( fd = 0; fd < 256; fd++ ){
		if( lastv+8 < fd )
			break;
		if( file_is(fd) ){
			sprintf(vp,"[%2d]",fd);
			vp += strlen(vp);
			nvfd++;
			lastv = fd;
		}
	}
	return nvfd;
}
void valid_fds(PCStr(where))
{	CStr(vfd,256);

	fprintf(stderr,"##[%d][%s]##",getpid(),where);
	valid_fdl(AVStr(vfd));
	fprintf(stderr,"%s\n",vfd);
}

static int touch_ctime(int fd){
	struct stat st0;
	int omode,xmode;

	if( fstat(fd,&st0) != 0 )
		return -1;

	omode = st0.st_mode;
	xmode = (omode & ~S_IEXEC) | (~(omode & S_IEXEC) & S_IEXEC);
	if( fchmod(fd,xmode) != 0 )
		return -1;
	return fchmod(fd,omode);
}

int fileIsremote1(int fd)
{	struct stat st1;
	int time0,time1;

	if( isatty(fd) )
		return 0;

	time0 = time(0);
	if( touch_ctime(fd) < 0 )
		return -1;
	time1 = time(0);
	if( fstat(fd,&st1) != 0 )
		return -1;

	if( st1.st_ctime < time0 || time1 < st1.st_ctime )
		return 1;
	else	return 0;
}

/* 
 * ctime is not updated for TMPFS files on SunOS
 * atime is not updated for NFS files
 */
int file_timeoff(int fd,int created_now)
{	struct stat st0,st1;
	int now;

	if( !file_isreg(fd) )
		return 0;

	if( fstat(fd,&st0) != 0 )
		return 0;

	if( touch_ctime(fd) != 0 )
		return 0;
	now = time(0);

	if( fstat(fd,&st1) != 0 )
		return 0;

	if( !created_now )
	if( st1.st_ctime == st0.st_ctime ){
		/* this will not work right after REAL CHANGE */
		return 0;
	}

	return now - st1.st_ctime;
}

int file_ctime(int fd)
{	struct stat stat;

	if( fstat(fd,&stat) == 0 )
		return stat.st_ctime;
	else	return -1;
}
int file_mtime(int fd)
{	struct stat stat;

	if( fstat(fd,&stat) == 0 )
		return stat.st_mtime;
	else	return -1;
}
int file_atime(int fd)
{	struct stat stat;

	if( fstat(fd,&stat) == 0 )
		return stat.st_atime;
	else	return -1;
}
int file_times(int fd,int *ctime,int *mtime,int *atime)
{	struct stat stat;

	if( fstat(fd,&stat) == 0 ){
		*ctime = stat.st_ctime;
		*mtime = stat.st_mtime;
		*atime = stat.st_atime;
		return 0;
	}else	return -1;
}
int File_sizetime(PCStr(path),int *fsize,int *ctime,int *mtime,int *atime)
{	struct stat st;

	if( stat(path,&st) == 0 ){
		if( fsize ) *fsize = st.st_size;
		if( ctime ) *ctime = st.st_ctime;
		if( mtime ) *mtime = st.st_mtime;
		if( atime ) *atime = st.st_atime;
		return 0;
	}else	return -1;
}
int file_sizetime(int fd,int *fsize,int *ctime,int *mtime,int *atime)
{	struct stat stat;

	if( fstat(fd,&stat) == 0 ){
		if( fsize ) *fsize = stat.st_size;
		if( ctime ) *ctime = stat.st_ctime;
		if( mtime ) *mtime = stat.st_mtime;
		if( atime ) *atime = stat.st_atime;
		return 0;
	}else	return -1;
}
int file_size(int fd)
{	struct stat stat;

	if( fstat(fd,&stat) == 0 )
		return stat.st_size;
	else	return -1;
}
int file_ino(int fd)
{	struct stat stat;

	if( fstat(fd,&stat) == 0 )
		return stat.st_ino;
	else	return 0;
}
int file_nlink(int fd)
{	struct stat stat;

	if( fstat(fd,&stat) == 0 )
		return stat.st_nlink;
	else	return -1;
}
int file_uid(int fd)
{	struct stat stat;

	if( fstat(fd,&stat) == 0 )
		return stat.st_uid;
	else	return -1;
}
int file_statX(int fd,int *sizp,int *mtmp,int *atmp,int *uidp,int *ftype)
{	struct stat stat;

	if( fstat(fd,&stat) == 0 ){
		if( sizp ) *sizp = stat.st_size;
		if( mtmp ) *mtmp = stat.st_mtime;
		if( atmp ) *atmp = stat.st_atime;
		if( uidp ) *uidp = stat.st_uid;
		if( ftype ){
			*ftype = '?';
			if( S_ISREG(stat.st_mode) ) *ftype = '-'; else
			if( S_ISDIR(stat.st_mode) ) *ftype = 'd'; else
			if( S_ISLNK(stat.st_mode) ) *ftype = 'l'; else
			if( S_ISFIFO(stat.st_mode) ) *ftype = 'p';
		}
		return 0;
	}else	return -1;
}
int file_stat(int fd,int *sizep,int *mtimep,int *atimep)
{	struct stat stat;

	if( fstat(fd,&stat) == 0 ){
		*sizep = stat.st_size;
		*mtimep = stat.st_mtime;
		*atimep = stat.st_atime;
		return 0;
	}else	return -1;
}
int file_copymod(int src,int dst)
{	struct stat stat;

	if( fstat(src,&stat) == 0 ){
		return fchmod(dst,stat.st_mode);
	}
	return -1;
}
int File_copymod(PCStr(src),PCStr(dst))
{	struct stat st;

	if( stat(src,&st) == 0 ){
		if( chmod(dst,st.st_mode) != 0 )
			return -1;
		if( set_utimes(dst,st.st_atime,st.st_mtime) != 0 )
			return -1;
		return 0;
	}
	return -1;
}

int file_isfifo(int fd)
{	struct stat status;

	if( fstat(fd,&status) == 0 )
		return S_ISFIFO(status.st_mode);
	return 0;
}

int file_device(int fd)
{	struct stat status;

	if( fstat(fd,&status) == 0 )
		return ((status.st_dev) >> 8) & 0xFF;
	return -1;
}
int File_device(PCStr(path))
{	struct stat status;

	if( stat(path,&status) == 0 )
		return ((status.st_dev) >> 8) & 0xFF;
	return -1;
}

int File_stat(PCStr(path),int *size,int *time,int *isdir)
{	struct stat status;

	if( stat(path,&status) == 0 ){
		*size = status.st_size;
		*time = status.st_mtime;
		if( S_ISDIR(status.st_mode) )
			*isdir = 1;
		else	*isdir = 0;
		return 0;
	}
	*size = -1;
	*time = -1;
	*isdir = 0;
	return -1;
}

void dump_file_stat(int fd)
{	struct stat stat;

	if( fstat(fd,&stat) != 0 )
		return;
	fprintf(stderr,"size: %d\n",(int)stat.st_size);
	fprintf(stderr,"mode: 0%o\n",stat.st_mode);
	fprintf(stderr,"uid: %d\n",stat.st_uid);
	fprintf(stderr,"gid: %d\n",stat.st_gid);
	fprintf(stderr,"atime: %d\n",stat.st_atime);
	fprintf(stderr,"mtime: %d\n",stat.st_mtime);
	fprintf(stderr,"ctime: %d\n",stat.st_ctime);
}

/*
 *	UNIX LS COMMAND
 */
char *FileModes(int mode){
	static char *smode; /**/
	int maski;

	if( smode == NULL )
		smode = (char*)StructAlloc(10);

	for( maski = 0; maski < 9; maski++ ){
		if( (1 << maski) & mode )
			smode[8-maski] = "xwrxwrxwr"[maski];
		else	smode[8-maski] = '-';
	}
	return smode;
}

#ifdef S_IFBLK
/*
#define Fkbytes(st)	(st.st_blocks/2)
*/
#define Fkbytes(st)	((int)((1<st.st_blocks)?(st.st_blocks/2):(st.st_size/1024)))
int stat_blocks(struct stat *stp){
	return stp->st_blocks;
}
#else
#define Fkbytes(st)	((st.st_size+1023)/1024)
int stat_blocks(struct stat *stp){
	return (stp->st_size / 0x8000 + 1) * 64;
}
#endif

#define Ftypec(st)	(S_ISDIR(st.st_mode)?'d':(S_ISLNK(st.st_mode)?'l':'-'))
#define Fmodes(st)	FileModes(st.st_mode)
#define Fnlink(st)	st.st_nlink
#define Fowner(st,b)	getUsernameCached(st.st_uid,b)
/*
#define Fbytes(st)	st.st_size
*/
#define Fbytes(st)	((int)st.st_size)
/*
#define Ftime(t,b)	rsctime(t,b)
#define Fmtime(st,b)	Ftime(st.st_mtime,b)
#define Fatime(st,b)	Ftime(st.st_atime,b)
*/

typedef struct {
	int	 s_ikey;
  const	char	*s_skey;
  const	char	*s_line;
} Elem;

typedef struct {
	FILE	*l_out;
  const char	*l_dir;
  const	char	*l_vbase;	/* put with specified virtual base directory */
 struct fa_stat	*l_maskrexp;
struct stat	*l_stp;
  const	char	*l_fmt;
	int	 l_all;
	int	 l_reflink;

	int	 l_ikey;
	defQStr(l_buf); /* to be used by AVStr */

	int	 l_nsize;
	int	 l_nfill;
	Elem	*l_lines;
} LsArg;

static int ls1(PCStr(file),LsArg *lsa)
{	const char *dir = lsa->l_dir;
	const char *fmt = lsa->l_fmt;
	const char *oline = lsa->l_buf;
	const char *sp;
	char sc;
	struct stat st;
	int rcode;
	int st_ok;
	int width;
	CStr(path,1024);
	CStr(buf,1024);
	refQStr(op,lsa->l_buf); /**/
	const char *tail;

	((char*)oline)[0] = 0;
	lsa->l_ikey = 0;
	if( file[0] == '.' && !lsa->l_all )
		return 0;

	if( lsa->l_maskrexp ){
		tail = frex_match(lsa->l_maskrexp,file);
		if( tail == NULL || *tail != 0 )
			return 0;
	}

	st_ok = 0;
	path[0] = 0;
	if( dir && dir[0] )
		strcpy(path,dir);
	if( *path )
	if( strtailchr(path) != '/' )
	if( *file != '/' )
		strcat(path,"/");
	strcat(path,file);

	if( lsa->l_stp ){
		st = *lsa->l_stp;
		st_ok = 1;
	}

	op = (char*)oline;
	for( sp = fmt; sc = *sp; sp++ ){
		assertVStr(lsa->l_buf,op+1);
		if( op[0] != 0 )
			op += strlen(op);

		if( sc != '%' || sp[1] == 0 ){
			setVStrPtrInc(op,sc); setVStrEnd(op,0);
			continue;
		}
		sp++;
		width = numscan(&sp);
		sc = *sp;

		if( sc == 'V' ){
			const char *rp;
			if( dir == 0 || dir[0] == 0 ){
				strcpy(op,lsa->l_vbase);
				continue;
			}
			strcpy(op,lsa->l_vbase);
			rp = strrpbrk(file,"/\\");
			if( rp && rp != file && rp[1] )
				rp += 1;
			else	rp = file;
			if( *op && strtailchr(op) != '/' && *rp != '/' )
				strcat(op,"/");
			strcat(op,rp);
			continue;
		}
		if( sc == 'N' ){
			strcpy(op,file);
			continue;
		}
		if( sc == 'A' ){
			strcpy(op,path);
			continue;
		}
		if( st_ok == 0 ){
			if( lsa->l_reflink || !INHERENT_lstat() )
				rcode = stat(path,&st);
			else	rcode = lstat(path,&st);
			if( rcode != 0 ){
				((char*)oline)[0] = 0;
				break;
			}
			st_ok = 1;
		}
		switch( sc ){
			case 'm': lsa->l_ikey = st.st_mtime; break;
			case 'a': lsa->l_ikey = st.st_atime; break;
			case 'z': lsa->l_ikey = st.st_size; break;
			case 'T': setVStrPtrInc(op,Ftypec(st)); setVStrEnd(op,0); break;
			case 'M': strcpy(op,Fmodes(st)); break;
			case 'L': sprintf(op,"%*d",width,Fnlink(st)); break;
			case 'O': sprintf(op,"%*s",width,Fowner(st,AVStr(buf))); break;
			case 'S': sprintf(op,"%*d",width,Fbytes(st)); break;
			case 'K': sprintf(op,"%*d",width,Fkbytes(st)); break;
			case 'D': sprintf(op,"%*s",width,rsctime(st.st_mtime,AVStr(buf))); break;
			case 'U': sprintf(op,"%*s",width,rsctime(st.st_atime,AVStr(buf))); break;
		}
	}
	return 0;
}

static scanDirFunc ls2(PCStr(file),LsArg *lsa)
{	int li,osize,nsize;

	ls1(file,lsa);
	if( lsa->l_buf[0] ){
		if( lsa->l_nfill == lsa->l_nsize ){
			osize = lsa->l_nsize;
			nsize = (lsa->l_nsize += 2048) * sizeof(Elem);
			if( osize == 0 )
				lsa->l_lines=(Elem*)malloc(nsize);
			else	lsa->l_lines=(Elem*)realloc(lsa->l_lines,nsize);
		}
		li = lsa->l_nfill++;
		lsa->l_lines[li].s_ikey = lsa->l_ikey;
		lsa->l_lines[li].s_skey = stralloc(file);
		lsa->l_lines[li].s_line = stralloc(lsa->l_buf);
	}
	return 0;
}
static int cmpline(Elem *e1,Elem *e2)
{	int diff;

	if( diff = e2->s_ikey - e1->s_ikey )
		return diff;
	else	return strcmp(e1->s_skey,e2->s_skey);
}
static void sort_ls(PCStr(dirpath),LsArg *lsa)
{	int li;
	const char *line;

	lsa->l_nfill = 0;
	lsa->l_nsize = 0;
	lsa->l_lines = NULL;

	Scandir(dirpath,scanDirCall ls2,lsa);
	qsort(lsa->l_lines,lsa->l_nfill,sizeof(Elem),(sortFunc)cmpline);

	for( li = 0; li < lsa->l_nfill; li++ ){
		line = lsa->l_lines[li].s_line;
		fprintf(lsa->l_out,"%s\r\n",line);
		free((char*)line);
		free((char*)lsa->l_lines[li].s_skey);
	}
	free(lsa->l_lines);
}

#define O_PUTSELF	1
#define O_TIMESORT	2
#define O_BYATIME	4
#define O_SIZESORT	8
#define O_FORM_L	0x10
#define O_FORM_S	0x20
#define O_REXP		0x40
#define O_VBASE		0x80

extern int TIME_NOW;
static int DoFileRexp;

void dir2ls(PCStr(dirpath),struct stat *stp,PCStr(opt),xPVStr(fmt),FILE *fp)
{	const char *op;
	const char *dp;
	int onow;
	int flags;
	char sortspec;
	LsArg lsa;
	CStr(fmt_s,256);
	CStr(fmt_a,256);
	CStr(line,2048);
	int f_l,f_s;
	CStr(fmt_b,256);
	CStr(spath,1024);
	const char *vbase = 0;
	CStr(vbaseb,1024);

	if( opt == NULL )
		opt = "";

	bzero(&lsa,sizeof(lsa));
	flags = 0;

	for( op = opt; *op; op++ ){
		switch( *op ){
			case 'a':
				lsa.l_all = 1;
				break;
			case 'l':
				flags |= O_FORM_L;
				break;
			case 's':
				flags |= O_FORM_S;
				break;
			case 'L':
				lsa.l_reflink = 1;
				break;
			case 'd':
				flags |= O_PUTSELF;
				break;
			case 't':
				flags |= O_TIMESORT;
				break;
			case 'z':
				flags |= O_SIZESORT;
				break;
			case 'u':
				flags |= O_BYATIME;
				break;
			case '*':
				flags |= O_REXP;
				break;
			case '/':
				flags |= O_VBASE;
				vbase = op + 1;
				while( op[1] ) *op++; /* skip remaining */
				if( strpbrk(vbase,"*?[") ){
/* vbase is "const" */
					strcpy(vbaseb,vbase);
					vbase = vbaseb;
					if( dp = strrchr(vbase,'/') )
						((char*)dp)[1] = 0;
					else	Xstrcpy(QVStr((char*)vbase,vbaseb),"");
				}
				break;
		}
	}
	if( (DoFileRexp || (flags & O_REXP)) && strpbrk(dirpath,"*?[") ){
		strcpy(spath,dirpath);
		dirpath = spath;
		if( dp = strrchr(dirpath,'/') ){
			lsa.l_maskrexp = frex_create(dp + 1);
			truncVStr(dp);
		}else{
			lsa.l_maskrexp = frex_create(dirpath);
			dirpath = ".";
		}
	}

	if( fmt == NULL ){
		cpyQStr(fmt,fmt_b);
		setVStrEnd(fmt,0);

		if( flags & O_FORM_S )
			strcat(fmt,"%4K ");
		if( flags & O_FORM_L )
			strcat(fmt,"%T%M%3L %-8O %8S %D ");

		if( flags & O_VBASE )
			strcat(fmt,"%V");
		else	strcat(fmt,"%N");
	}

	if( flags & O_TIMESORT ){
		if( flags & O_BYATIME )
			sortspec = 'a';
		else	sortspec = 'm';
		sprintf(fmt_s,"%%%c%s",sortspec,fmt);
		fmt = fmt_s;
	}
	if( flags & O_SIZESORT ){
		sprintf(fmt_s,"%%%c%s",'z',fmt);
		fmt = fmt_s;
	}
	if( flags & O_BYATIME ){
		strcpy(fmt_a,fmt);
		fmt = fmt_a;
		if( dp = strstr(fmt,"%D") )
			((char*)dp)[1] = 'U';
	}

	lsa.l_dir = dirpath;
	lsa.l_vbase = (char*)vbase;
	lsa.l_stp = stp;
	lsa.l_fmt = (char*)fmt;
	lsa.l_out = fp;
	setQStr(lsa.l_buf,line,sizeof(line));
	line[0] = 0;

	onow = TIME_NOW;
	TIME_NOW = time(0);
	setpwent();

	if( !(flags & O_PUTSELF) && fileIsdir(dirpath) )
		sort_ls(dirpath,&lsa);
	else{
		lsa.l_dir = "";
		lsa.l_all = 1;
		ls1(dirpath,&lsa);
		fprintf(fp,"%s\r\n",lsa.l_buf);
	}

	endpwent();
	TIME_NOW = onow;

	if( lsa.l_maskrexp )
		frex_free(lsa.l_maskrexp);
	return;
}

FILE *ls_unix(FILE *fp,PCStr(opt),PVStr(fmt),PCStr(dir),struct stat *stp)
{	int io[2];

	if( dir == NULL )
		dir = ".";

	if( fp != NULL ){
		dir2ls(dir,stp,opt,BVStr(fmt),fp);
		return fp;
	}else{
		pipe(io);
		if( fork() == 0 ){
			close(io[0]);
			fp = fdopen(io[1],"w");
			dir2ls(dir,stp,opt,BVStr(fmt),fp);
			exit(0);
			return NULL;
		}else{
			close(io[1]);
			fp = fdopen(io[0],"r");
			return fp;
		}
	}
}

int ls_main(int ac,const char *av[])
{	const char *fmt;

	fmt = getenv("LSFMT");
	if( ac <= 1 )
		ls_unix(stdout,NULL,CVStr(fmt),".",NULL);
	else
	if( av[1][0] == '-' )
		ls_unix(stdout,av[1],CVStr(fmt),av[2],NULL);
	else	ls_unix(stdout,NULL,CVStr(fmt),av[1],NULL);
	return 0;
}
#ifdef LSMAIN
main(ac,av) char *av[]; { DoFileRexp = 1; ls_main(ac,av); }
#endif

static scanDirFunc fmatch(PCStr(file),PCStr(pat),int *nump)
{
	if( strstr(file,pat) != 0 ){
		*nump += 1;
	}
	return 0;
}
int file_matches(PCStr(dir),PCStr(pat))
{	int num;

	num = 0;
	Scandir(dir,scanDirCall fmatch,pat,&num);
	return num;
}
