/*////////////////////////////////////////////////////////////////////////
Copyright (c) 2005 National Institute of Advanced Industrial Science and Technology (AIST)

Permission to use this material for evaluation, copy this material for
your own use, and distribute the copies via publically accessible on-line
media, without fee, is hereby granted provided that the above copyright
notice and this permission notice appear in all copies.
AIST MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	httpccx.c (character code conversion on HTTP)
Author:		Yutaka Sato <ysato@delegate.org>
Description:
    TODO
	counter for mixed charset server:
	  SVCC=shift_JIS/12,euc-jp=/34,...
	see SVCC when without explicit charset and not obviousJP()
History:
	050714	extracted from http.c
//////////////////////////////////////////////////////////////////////#*/
#include "delegate.h"
#include "http.h"
#include "file.h"


#define getFV(str,fld,buf)  getFieldValue2(str,fld,AVStr(buf),sizeof(buf))
int getParam(PVStr(params),PCStr(name),PVStr(val1),int siz,int del);
extern const char *TIMEFORM_COOKIE;

/*
 * SVCC is a hint for the charset of texts which is acceptable by
 * the server.  It is created from the charset of server's response.
 * It is sent from client in a Cookie in a request to indicate to
 * which charset a request to the server should be converted.
 */
#define DGC_SVCC "DeleGate-Control-SVCC"
#define NOCONV	"NoConv"
static char SVCCcookie[32]; /* SVCC sent from client in Cookie */
static int SVCCnoConv; /* SVCC=NoConv to disalbe CCX */
static char OutCharset[32];

void HTCCX_init(){
	SVCCcookie[0] = 0;
	SVCCnoConv = 0;
	OutCharset[0] = 0;
}

void HTCCX_putSVCC(Connection *Conn,FILE *tc,int rcode,PCStr(head)){
	const char *xcharset;
	CStr(svcc,64);

	if( CCXactive(CCX_TOCL) == 0 )
		return;

	svcc[0] = 0;
	if( InCharset[0] == 0 )
	{
		const char *xch;
		CStr(rcs,64); /* peeked charset */
		CStr(xhead,512);
		CStr(tmp,1024);
		rcs[0] = 0;
		strcpy(xhead,head);
		extractParam(AVStr(xhead),"Content-Type","charset",AVStr(rcs),
			sizeof(rcs),0);
/*
CCXlog("{C} head charset[%s] In[%s]\n",rcs,InCharset);
*/
		if( rcs[0] == 0 ){
		/* should try scan_metahttp "HTTP-EQUIV Content-Type charset" */
		CCXexec(CCX_TOCL,head,strlen(head),AVStr(tmp),sizeof(tmp));
		xch = CCXident(CCX_TOCL); 
if( xch )
if( !strcaseeq(xch,"US-ASCII") )
CCXlog("{C} ---- [%s] head guessed[%s]\n",InCharset,xch?xch:"?");
		}
	}

	if( svcc[0] == 0 && InCharset[0] == 0 && CCX_converror(CCX_TOCL) == 0 )
		return;

	if( CCX_converror(CCX_TOCL) ){
		/* conversion is failed */
		CCXlog("{C} conv-error sv[%s] In[%s] SV[%s]\n",
			svcc,InCharset,SVCCcookie);
		if( InCharset[0] && !streq(InCharset,SVCCcookie) ){
			/* conv. req. by SVCC is not tried yet */
			xcharset = InCharset;
		}else{
			/*
			xcharset = NOCONV;
			*/
			xcharset = InCharset;
		}
	}else{
		xcharset = svcc;
		if( xcharset[0] == 0 )
			/* xcharset = CCXident(CCX_TOCL); */
			xcharset = InCharset;
		if( xcharset == NULL )
			xcharset = "";
	}
	if( 300 <= rcode ){
		/* error message might in non-localized charset */
		CCXlog("{C} set SVCC[%s] cancelled by code=%d\n",
			xcharset,rcode);
		return;
	}
	if( !strcaseeq(xcharset,"US-ASCII") ){
		CStr(params,1024);
		CStr(path,1024);
		CStr(ex,64);
		int exp;
		/* the path should be the MOUNT point ... */
/*
		exp = time(0) + 10*60;
		StrftimeGMT(AVStr(ex),sizeof(ex),TIMEFORM_COOKIE,exp,0);
		sprintf(params,"%s=%s; Expires=%s",DGC_SVCC,xcharset,ex);
*/
/*
		sprintf(params,"%s=%s",DGC_SVCC,xcharset);
		fprintf(tc,"Set-Cookie: %s\r\n",params);
*/

		HTTP_originalURLPath(Conn,AVStr(path));
		if( *path ){
			const char *dp,*np;
			if( dp = strpbrk(path,"=+&%") )
				truncVStr(dp);
			if( dp = strchr(path,'?') )
				truncVStr(dp);
			else
			if( dp = strchr(path+1,'/') ){
				if( np = strchr(dp+1,'/') )
					truncVStr(np);
				else	truncVStr(dp);
			}
			else{
				if( path[0] == '/' )
					path[1] = 0;
			}
		}
		if( *path && strcmp(path,"/") != 0 ){
			sprintf(params,"%s=%s",DGC_SVCC,xcharset);
			fprintf(tc,"Set-Cookie: %s; Path=%s\r\n",params,path);
		}
		CCXlog("{C}*set SVCC[%s][%s]%s\n",xcharset,DST_HOST,path);

		exp = time(0) + 24*60*60;
		strcpy(path,"/");
		StrftimeGMT(AVStr(ex),sizeof(ex),TIMEFORM_COOKIE,exp,0);
		sprintf(params,"%s=%s; Expires=%s",DGC_SVCC,xcharset,ex);
		fprintf(tc,"Set-Cookie: %s; Path=%s\r\n",params,path);

		/*
		if( no Vary charset ){
			saveSVCCcache(Conn);
		}
		*/
	}
}

void HTCCX_getSVCC(Connection *Conn,PVStr(f)){
	CStr(cc,128);

	if( !CCXactive(CCX_TOCL) && !CCXactive(CCX_TOSV) ){
		/* SVCC is not hop-by-hop */
		return;
	}

	if( extractParam(AVStr(f),"Cookie",DGC_SVCC,AVStr(cc),sizeof(cc),1) ){
		if( cc[0] ){
			Xstrcpy(FVStr(SVCCcookie),cc);
		}
	}
}

static int saveSVCCcache(Connection *Conn){
	return 0;
}
static int loadSVCCcache(Connection *Conn,PCStr(proto),PCStr(site),int port,PCStr(upath)){
	CStr(cpath,1024);
	CCXlog("----A {C} SVCC cache ? %s [%s][%s]%d[%s] cachedir=%X\n",REQ_URL,
		proto,site,port,upath,cachedir());
	if( CTX_cache_path(Conn,proto,site,port,upath,AVStr(cpath)) == 0 )
		return 0;
	CCXlog("----B {C} SVCC cache ? %s [%s][%s]%d[%s] %s\n",REQ_URL,
		proto,site,port,upath,cpath);
	return 0;
}

static void initCCXbySVCC(Connection *Conn,PCStr(proto),PCStr(site),int port,PCStr(upath)){
	const char *stat = "";

	if( CCXactive(CCX_TOSV) != 0 )
		return;

	if( SVCCcookie[0] == 0 ){
/*
CCXlog("---- {C} CCX_TOSV=%X CCX_TOCL=%X SVCC=[%s]\n",
CCXactive(CCX_TOSV),CCXactive(CCX_TOCL),SVCCcookie);
		if( CCXactive(CCX_TOCL) == 0
		 || loadSVCCcache(Conn,proto,site,port,upath) == 0 )
*/
			return;
	}

	if( 2 <= HttpReload ){
		stat = "IGNORED";
	}else
	if( streq(SVCCcookie,NOCONV) ){
		SVCCnoConv = 1;
	}else{
		CCXcreate("*",SVCCcookie,CCX_TOSV);
		/* also it could be a hint for the input from server
		 * when it is ambiguous
		 * CCX_setincode(CCX_TOCL,SVCCcookie)
		 */
	}

	if( lCHARSET() )
	if( fileSeemsBinary(REQ_URL) <= 0 )
	{
		CStr(dom,32);
		refQStr(dp,dom);
		if( strncmp(REQ_URL,"http://",7) == 0 )
			FStrncpy(dom,REQ_URL+7);
		else	FStrncpy(dom,REQ_URL);
		if( dp = strchr(dom,'/') )
		if( dp = strchr(dp+1,'/') )
			setVStrEnd(dp,1);
		CCXlog("{C}*got SVCC[%s][%s]%s\n",SVCCcookie,dom,stat);
	}
}

void HTCCX_setInputCode(Connection *Conn,PCStr(where),CCXP ccx,PCStr(cset)){
	Xstrcpy(FVStr(InCharset),cset);
	CCXlog("{C} %s InCharset[%s] %d\n",where,InCharset,CCXwithJP(ccx));

	/* set the charset into CCX as the hint on input-charset */
	/*if( CCXwithJP(ccx) == 0 ) should be overridden HEAD < XML < META */
	if( cset[0] && CCXcharsetcmp(cset,"ISO-2022-JP") != 0 ){
		sv1log("{C} set CCX in-charset[%s] by %s\n",cset,where);
		CCX_setincode(ccx,cset);
	}
}

static int thruCode(Connection *Conn,PCStr(where),CCXP ccx,PCStr(cset)){
	const char *ccxout = 0;

	HTCCX_setInputCode(Conn,where,ccx,cset);

	/* don't erase charset=... if the charset is not changed */
	if( CCXactive(ccx) )
	if( CCXoutcharset(ccx,&ccxout) )
	if( ccxout != 0 && CCXcharsetcmp(ccxout,cset) == 0 ){
		sv1log("{C} thru chaset=%s in %s\n",cset,where);
		return 1;
	}

	if( CCXguessing(ccx) ){ /* CHARSET=guess */
		sv1log("{C} thru chaset=%s in %s (CHARSET=guess)\n",cset,where);
		Xstrcpy(FVStr(InCharset),cset);
		return 1;
	}

	if( CCX_inputIsJP(ccx) == 0 ){
		sv1log("{C} thru nonJP chaset=%s in %s\n",cset,where);
		return 1;
	}
	return 0;
}

void HTCCX_setReqSVCC(Connection *Conn,PCStr(proto),PCStr(site),int port,PCStr(upath)){
	if( HttpReload ){
		if( SVCCcookie[0] )
		if( fileSeemsBinary(REQ_URL) <= 0 )
		{
			CCXlog("{C} %sSVCC[%s] with reload-%d [%s]\n",
				2<=HttpReload?"(ignore)":"",
				SVCCcookie,HttpReload,REQ_URL);
		}
		if( 2 <= HttpReload ){
			SVCCcookie[0] = 0;
			SVCCnoConv = 0;
		}
	}
	//if( SVCCcookie[0] )
	{
		initCCXbySVCC(Conn,proto,site,port,upath);
	}
}

int HTCCX_replaceCharset(Connection *Conn,PVStr(field),int fnlen,FILE *fp){
	const char *xcharset;
	CStr(ctype,64);
	CStr(chset,32);

	Xsscanf(field+fnlen," %[a-zA-Z0-9/-]",AVStr(ctype));
	if( !STRH(field,F_CtypeText)
	 && !streq(ctype,"application/x-javascript")
	 && !streq(ctype,"application/xml")
	){
		return 0;
	}
	getParam(DVStr(field,fnlen),"charset",AVStr(chset),sizeof(chset),0);

	if( xcharset = HTTP_outCharset(Conn) )
	if( CCXactive(CCX_TOCL) )
	{
		if( chset[0] != 0 )
			HTCCX_setInputCode(Conn,"HEAD",CCX_TOCL,chset);

		if( chset[0] == 0 )
		if( file_isreg(fileno(fp)) )
		{
			int off;
			int rcc;
			char buff[4096];
			CStr(ob,4096);
			CStr(st,64);

			off = ftell(fp);
			rcc = fread(buff,1,sizeof(buff)-1,fp);
			fseek(fp,off,0);
			if( 0 < rcc ){
				buff[rcc] = 0;
				CCXexec(CCX_TOCL,buff,rcc,AVStr(ob),sizeof(ob));
			}
			CCXstats(CCX_TOCL,AVStr(st));
			CCXlog("{C} [%s] nonJP=%d %s\n",ctype,
				CCXnonJP(CCX_TOCL),st);
		}
		if( CCXnonJP(CCX_TOCL) ){
			CCXlog("{C} Head thru nonJP[%s] %s",chset,field);
		}else
		if( SVCCnoConv == 0 ){
			replace_charset(AVStr(field),xcharset);
			CCXlog("{C} Head replaced[%s] %s",chset,field);
			Xstrcpy(FVStr(OutCharset),xcharset);
			return 1;
		}
	}
	return 0;
}

int CCX_obviousJP(CCXP ccx);
int HTCCX_restoreCharset(Connection *Conn,PVStr(field),int fnlen,PCStr(where)){
	CStr(buf,128);
	CStr(ctype,64);
	CStr(cs,64);
	int nonJP = 0;

	if( Conn == NULL )
		return 0;

	Xsscanf(field+fnlen," %[a-zA-Z0-9/-]",AVStr(ctype));
	if( !STRH(field,F_CtypeText)
	 && !streq(ctype,"application/x-javascript")
	 && !streq(ctype,"application/xml")
	){
		return 0;
	}
	getParam(DVStr(field,fnlen),"charset",AVStr(cs),sizeof(cs),0);

	CCXstats(CCX_TOCL,AVStr(buf));
	CCXlog("{C} %s[%s] SVCC[%s] %s%s%s %s %s",where,InCharset,SVCCcookie,
		CCXactive(CCX_TOCL)?"A":"-",
		CCXwithJP(CCX_TOCL)?"J":"-",
		CCXnonJP(CCX_TOCL)?"?":"-",
		buf,field);

	if( CCXactive(CCX_TOCL) == 0 )
		return 0;

	if( CCXguessing(CCX_TOCL) ){
		/* setGuessedCharset() will be called after ... */
		return 0;
	}

	if( InCharset[0] == 0 )
	if( InLang[0] && !strcaseeq(InLang,"jp") )
	{
		if( CCX_obviousJP(CCX_TOCL) == 0 ){
			/*
			nonJP = 1;
			*/
			CCXlog("{C} nonJP Lang[%s] [%s]\n",InLang,InCharset);
		}
	}

	if( nonJP == 0 )
	if( CCXwithJP(CCX_TOCL) )
	if( CCXnonJP(CCX_TOCL) == 0 )
	{
		const char *outchset;
		if( OutCharset[0] != 0 )
			return 0;
		/* out-charset was not put in the header because
		 * unknown or non-JP charset name was given originally,
		 * but after then CCX charset was set by META and
		 * conversion was done.
		 */
		if( CCX_converror(CCX_TOCL) )
			return 0;
		/*
		if( CCXoutcharset(CCX_TOCL,&outchset) == 0 || outchset == 0 )
		*/
		if( (outchset = CCXcharset(CCX_TOCL)) == 0 )
			return 0;
		replace_charset(AVStr(field),outchset);
		CCXlog("{C} %s/Head replace[%s]<-[%s][%s] %s",
			where,outchset,cs,InCharset,field);
		/* for asahi-net */
		return 1;
	}

	/*
	strcpy(cs,"");
	*/
	if( InCharset[0] )
		replace_charset(AVStr(field),InCharset);
	else{
		getParam(DVStr(field,fnlen),"charset",AVStr(cs),sizeof(cs),1);
	}
	CCXlog("{C} %s[%s]<-[%s] restoreCharset %s",where,InCharset,cs,field);
	return 1;
}

int HTCCX_guessCharset(Connection *Conn,PCStr(ctype))
{ 
	if( strcasestr(ctype,"text/html") || strcasestr(ctype,"text/plain") )
	if( strcasestr(ctype,"charset=") == 0 )
	if( CCXactive(CCX_TOCL) )
	if( CCXguessing(CCX_TOCL) ) /* CHARSET=guess */
		return 1;
	return 0;
}

void HTCCX_setGuessedCharset(Connection *Conn,PVStr(line))
{	int fnlen;
	const char *xcharset;

	if( fnlen = STRH(line,F_ContType) )
	if( HTCCX_guessCharset(Conn,line+fnlen) ){
		if( xcharset = CCXident(CCX_TOCL) ){
			if( strcaseeq(xcharset,"US-ASCII") )
				xcharset = InCharset;
			if( *xcharset )
			if( !strcaseeq(xcharset,"US-ASCII") ){
				replace_charset(AVStr(line),xcharset);
				sv1log("{C} charset guessed:%s",line+fnlen);
			}
		}
	}
}

void HTCCX_reqHead(Connection *Conn)
{	const char *method;
	const char *ver;
	HttpRequest reqx;
	CStr(url1,URLSZ);
	CStr(url2,URLSZ);
	refQStr(sp1,url1); /**/
	int oc;

	if( CCXactive(CCX_TOSV) == 0 )
		return;

	decomp_http_request(REQ,&reqx);
	method = reqx.hq_method;
	strcpy(url1,reqx.hq_url);
	ver = reqx.hq_ver;
	

	if( sp1 = strchr(url1,'?') ){
		strcpy(url2,sp1+1);
		sprintf(sp1,"-_-_-%s",url2);
	}

	/* should do CCX_setincode(CHARSET-param) because the client is
	 * expected to send request string in the response CHARSET
	 * if the request is not in the CHARSET, then start without
	 * the hypothesis.
	 */

	if( strchr(REQ,'%') ){
		CCXlog("{C} CCX_REQ [%s] %s\n",SVCCcookie,url1);
	}
	URL_unescape(url1,AVStr(url2),0,1);
	/*
	nonxalpha_unescape(url1,AVStr(url2),0);
	*/
	oc = CCXexec(CCX_TOSV,url2,strlen(url2),AVStr(url1),sizeof(url1));
	CCXexec(CCX_TOSV,"",0,QVStr(url1+oc,url1),sizeof(url1)-oc);
	if( strchr(REQ,'%') ){
		CStr(st,128);
		CCXstats(CCX_TOSV,AVStr(st));
		CCXlog("{C} CCX_REQ convstat=%s\n",st);
	}

	if( strcmp(url1,url2) != 0 ){
		refQStr(sp2,url2);
		URL_reescape(url1,AVStr(url2),0,1);
		/*
		nonxalpha_escapeX(url1,AVStr(url2),sizeof(url2));
		*/
		if( sp2 = strstr(url2,"-_-_-") )
			sprintf(sp2,"?%s",sp2+5);
		sprintf(url1,"%s %s HTTP/%s\r\n",method,url2,ver);
		sv1log("{C} applied char. conv. for request header (%d<-%d)\n",
			strlen(url1),strlen(REQ));
		CCXlog("{C} CCX_REQ [%s] (%d <- %d) %s\n",SVCCcookie,
			strlen(url1),strlen(REQ),url2);
		strcpy(REQ,url1);
		Verbose(">> %s",REQ);
	}
}

int HTCCX_reqBody(Connection *Conn,FILE *ts,PCStr(req),PCStr(fields),PCStr(body),int blen){
	CStr(ctype,128);
	CStr(cleng,32);
	defQStr(xbody);
	defQStr(ybody);
	defQStr(xfields);
	int bsiz;
	int hsiz;
	int oc;
	const char *cs;

	if( getFV(fields,"Content-Type",ctype) ){
	  if( CCXactive(CCX_TOSV) ){
	    if( streq(ctype,"application/x-www-form-urlencoded") ){
		bsiz = blen * 2;
		setQStr(xbody,malloc(bsiz),bsiz);
		setQStr(ybody,malloc(bsiz),bsiz);
		sv1log("{C} --qbody--URLESC-I %s\n",body);
		URL_unescape(body,AVStr(xbody),1,1);

		oc = CCXexec(CCX_TOSV,xbody,strlen(xbody),AVStr(ybody),bsiz);
		CCXexec(CCX_TOSV,"",0,QVStr(ybody+oc,ybody),bsiz-oc);
		URL_reescape(ybody,AVStr(xbody),1,1);
		sv1log("{C} --qbody--URLESC-O %s\n",xbody);
		sprintf(cleng,"%d",strlen(xbody));

		hsiz = strlen(fields) + 128;
		setQStr(xfields,malloc(hsiz),hsiz);
		strcpy(xfields,fields);
		replaceFieldValue(AVStr(xfields),"Content-Length",cleng);
		sv1log("{C} applied char. conv. for request body (%s<-%d)\n",
			cleng,blen);
		cs = CCXcharset(CCX_TOSV);
		if( cs == 0 )
			cs = "";
		sv1log("{C} CCX to[%s] from[%s]\n",cs,CCXident(CCX_TOSV));

		fputs(req,ts);
		fputs(xfields,ts);
		fputs(xbody,ts);
		return 1;
	    }
	  }
	}
	return 0;
}

/*
	if( CCXactive(CCX_TOSV) == 0 )
	if( CCXactive(CCX_TOCL) != 0 ){
		int tobeccx = 0;
		const char *qp;
		char ch;
		for( qp = REQ_URL; ch = *qp; qp++ ){
			if( (ch & 0x80) || ch == '%' ){
				tobeccx = 1;
				break;
			}
		}
		CCXcreate("*",outcs,CCX_TOSV);
		CCXsetincode(CCX_TOSV,incs);
	}
*/

static const char *metahttp(Connection *Conn,PCStr(attr),PVStr(cont),PCStr(tagp),PVStr(contp),PCStr(nextp))
{	CStr(cset,64);
	int len;

	if( strcaseeq(attr,"Content-Language") ){
		strcpy(InLang,cont);
		return nextp;
	}

	if( strncaseeq(tagp,"<?xml",5) )
	if( strcaseeq(attr,"encoding") ){
		int olen;
		if( thruCode(Conn,"?XML",CCX_TOCL,cont) )
			return nextp;
		olen = strlen(contp);
		getParam(BVStr(contp),"encoding",AVStr(cset),sizeof(cset),1);
		len = strlen(contp);
		sv1log("{C} removed charset in ?XML[%s][%s]\n",attr,cont);
		nextp = nextp - (olen-len);
		return nextp;
	}

	cset[0] = 0;
	if( strcaseeq(attr,"Content-Type") )
	if( len = erase_charset_param(AVStr(cont),AVStr(cset)) ){
		if( thruCode(Conn,"META",CCX_TOCL,cset) )
			return nextp;
		len = erase_charset_param(AVStr(contp),AVStr(cset));
		sv1log("{C} removed charset in HTTP-EQUIV[%s][%s]\n",cont,cset);
		return nextp - len;
	}
	return nextp;
}

typedef const char *(*scanMetaFuncP)(Connection *Conn,PCStr(attr),PVStr(cont),PCStr(tagp),PVStr(contp),PCStr(nextp));
void scan_metahttp(Connection*ctx,PVStr(line),scanMetaFuncP func);
static int htmlconv(Connection *Conn,PCStr(ctype),PVStr(Cline));

int HTCCX_html(Connection *Conn,PCStr(ctype),int size,PVStr(Cline),PVStr(Nline)){
	if( SVCCnoConv )
		return 0;

/*
	htmlconv(Conn,ctype,BVStr(Cline));
*/
	scan_metahttp(Conn,BVStr(Cline),metahttp);
	if( CCXactive(CCX_TOCL) )
		CCXexec(CCX_TOCL,Cline,strlen(Cline),BVStr(Nline),size);
	else	CTX_line_codeconv(Conn,Cline,BVStr(Nline),ctype);
	return 1;
}


/*
 * 050718 extracted from delegated.c
 */
#include "param.h"
static int loadCCXmap(PCStr(map),int mi,int mn){
	CStr(url,1024);
	CStr(path,1024);
	FILE *tmp,*fp;
	int loaded = 0;

	sprintf(url,"%s/codemap/%s",DELEGATE_homepage(),map);
	fprintf(stderr,"---- %d/%d downloading %s ...\n",mi,mn,url);
	sprintf(path,"${LIBDIR}/codemap/%s",map);
	Substfile(path);
	if( 0 < File_size(path) ){
		fprintf(stderr,"---- ALREADY at %s (%dbytes)\n",path,
			File_size(path));
		return 0;
	}

	tmp = TMPFILE("CCXtab");
	if( tmp == NULL )
		return 0;

	URLget(url,0,tmp); /* global CHARSET should be disabled ... */
	if( !feof(tmp) && 0 < file_size(fileno(tmp)) ){
		if( fp = dirfopen("CCXMAP",AVStr(path),"w") ){
			copyfile1(tmp,fp);
			fclose(fp);
			fprintf(stderr,"---- INSTALLED at %s (%dbytes)\n",path,
				File_size(path));
			loaded = 1;
		}
		else{
			fprintf(stderr,"???? FAILED to create %s\n",path);
		}
	}
	else{
		fprintf(stderr,"???? FAILED to download %s\n",url);
	}
	fclose(tmp);
	return loaded;
}

int CCX_JUmapCache(PCStr(jumap),PVStr(bpath));
int get_builtin_MADE_TIME();
void downloadCCXTabs(Connection *Conn)
{	const char *charset;
	int loaded = 0;

	if( DELEGATE_getEnv(P_CHARSET) || DELEGATE_getEnv(P_CHARCODE) )
	/*
	if( (charset = CCXcharset(CCX_TOCL)) && strcmp(charset,"UTF-8") == 0 ){
	*/
	if( CCXcharset(CCX_TOCL) && CCXoutJP(CCX_TOCL) ){
		CStr(path,1024);
		if( 0 < CCX_JUmapCache(NULL,AVStr(path)) )
		if( get_builtin_MADE_TIME() <= File_mtime(path) )
		if( 0 < UCSinit() )
			return;

		fprintf(stderr,"**** downloading CHARSET tables ....\n");
		loaded += loadCCXmap("JIS0208.TXT",1,5);
		loaded += loadCCXmap("JIS0212.TXT",2,5);
		loaded += loadCCXmap("SHIFTJIS.TXT",3,5);
		loaded += loadCCXmap("CP932.TXT",4,5);
		loaded += loadCCXmap("JAPANESE.TXT",5,5);
		fprintf(stderr,"**** downloaded %d tables\n",loaded);
		UCSreset();
		if( 0 < UCSinit() ){
			fprintf(stderr,"**** cached at %s\n",path);
			File_touch(path,time(0));
		}else{
			fprintf(stderr,"**** ERROR in the tables...\n");
		}
	}
}

typedef struct {
	char *src;
	char *dst;
} Subst;
static Subst subst[128];
static int substx;

static int htmlconv(Connection *Conn,PCStr(ctype),PVStr(line)){
	int i;
	if( substx == 0 ){
		FILE *fp;
		CStr(line,1024);
		CStr(src,256);
		CStr(dst,256);
		const char *dp;
		if( fp = fopen("mlb.cnv","r") ){
			while( fgets(line,sizeof(line),fp) != NULL ){
				dp = wordScanY(line,src,"^\t");
				if( *dp != '\t' )
					continue;
				while( *dp == '\t' )
					dp++;
				lineScan(dp,dst);
				if( *src && *dst ){
					subst[substx].src = stralloc(src);
					subst[substx].dst = stralloc(dst);
					substx++;
				}
			}
fprintf(stderr,"---- loaded %d\n",substx);
			fclose(fp);
		}else{
		}
	}
	for( i = 0; i < substx; i++ ){
		CStr(oline,0x10000);
		CStr(buf1,0x10000);
		CStr(buf2,0x10000);
		strcpy(oline,line);
		strsubst(BVStr(line),subst[i].src,subst[i].dst);

		if( strcmp(oline,line) != 0 ){
fprintf(stderr,"#A# [%s] %s %s\n",ctype,subst[i].src,REQ_URL);
if(strstr(line,"TINO"))
fprintf(stderr,"----------------------------------\n");
/*
fprintf(stderr,"#A# %s\n",line);
			URL_unescape(line,AVStr(buf1),0,1);
		CCXexec(CCX_TOCL,buf1,strlen(buf1),AVStr(buf2),sizeof(buf2));
			URL_reescape(buf2,AVStr(line),0,1);
fprintf(stderr,"#B# %s\n",line);
*/
		}
	}
	return 0;
}
