/*////////////////////////////////////////////////////////////////////////
Copyright (c) 1994-1999 Yutaka Sato
Copyright (c) 1994-1999 Electrotechnical Laboratry (ETL), AIST, MITI

Permission to use, copy, and distribute this material for any purpose
and without fee is hereby granted, provided that the above copyright
notice and this permission notice appear in all copies.
ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	iotimeout.c
Author:		Yutaka Sato <ysato@etl.go.jp>
Description:
History:
	940802	created

timeout could be the function of I/O buffer size...

//////////////////////////////////////////////////////////////////////#*/

#include <stdio.h>
#include "ystring.h"
#include "vsocket.h"
#include "vsignal.h"
#include "ysocket.h"
#include "fpoll.h"
#include "dglib.h"
#include "file.h"
#include <errno.h>

#define sv1log	syslog_ERROR

#define CHECK_ITVL	(10*1000)

extern int (*CheckClosed)(int);
int checkCloseOnTimeout(int checktime){
	if( CheckClosed )
		return (*CheckClosed)(checktime);
	else	return 0;
}

int xPollIn(int fd,int timeout)
{	int nready,ntry;
	int remain,timeout1;
	int closed;

	if( readyAlways(fd) )
		return 1;

	nready = -1;
	remain = timeout;
	closed = 0;

	for( ntry = 0; timeout == 0 && ntry < 1 || 0 < remain; ntry++ ){
		if( closed = checkCloseOnTimeout(1) )
			timeout1 = remain;
		else
		if( CHECK_ITVL < remain )
			timeout1 = CHECK_ITVL;
		else	timeout1 = remain;
		errno = 0;
		nready = PollIn(fd,timeout1);
		if( 0 < nready )
			break;

/*
this was introduced as a trial to reset possible POLL_PRI status
at DeleGate/5.7.5 (17Nov98), but seemed useless,
and was got rid of at DeleGate/5.9.1 because it is harmful; at least
it cause blocking on slow I/O on Win32.
		if( nready == 0 ){
			CStr(buf,1);
			if( RecvOOB(fd,buf,0) == 0 )
			sv1log("#### POLL_PRI ? PollIn()==0, RecvOOB()==0\n");
		}
*/

		if( nready < 0 ){
			if( errno == EINTR )
				sv1log("#### INTERRUPTED PollIn()=%d\n",nready);
			else{
				sv1log("#### xPollIn(%d,%d)=%d errno=%d\n",
					fd,timeout,nready,errno);
				break;
			}
		}else	remain -= timeout1;
	}
	if( closed == 0 )
		checkCloseOnTimeout(1);
	if( 100 < ntry )
		sv1log("#### xPollIn(%d,%d)*%d : %d\n",fd,timeout,ntry,nready);
	return nready;
}

/*
 * I/O timeout in seconds
 */
int IO_TIMEOUT = (10*60);

static sigjmp_buf ioenv;
static int io_fd;
static int io_timer;
static int io_sig;
static void (*io_pipe)(int sig);

static void io_timeout(int sig)
{
	io_sig = sig;
	siglongjmp(ioenv,sig);
}
#define	RETURN_ONTIMEOUTX(fd,retv,timeout) \
	checkCloseOnTimeout(1); \
	io_fd = fd; \
	io_pipe = Vsignal(SIGPIPE,io_timeout); \
	io_timer = pushTimer("Timered-IO",io_timeout,timeout); \
	if( sigsetjmp(ioenv,1) != 0 ){ \
		Vsignal(SIGPIPE,io_pipe); \
		popTimer(io_timer); \
		checkCloseOnTimeout(1); \
		sv1log("IO_TIMEOUT(%d) SIG%s\n",timeout,sigsym(io_sig)); \
		return retv; \
	}

#define	RETURN_ONTIMEOUT(fd,retv) RETURN_ONTIMEOUTX(fd,retv,IO_TIMEOUT)

#define DONE_SUCCESSFULLY() \
	Vsignal(SIGPIPE,io_pipe); \
	popTimer(io_timer); \
	checkCloseOnTimeout(1);

/*
int writevTimeout(int fd,char **iov,int nv,int timeout)
{	int wcc;

	RETURN_ONTIMEOUTX(fd,-1,timeout);
	wcc = writev(fd,iov,nv);
	DONE_SUCCESSFULLY();
	return wcc;
}
*/
int sendTimeout(int fd,PCStr(buff),int leng,int flag,int timeout)
{	int wcc;

	RETURN_ONTIMEOUTX(fd,-1,timeout);
	wcc = send(fd,buff,leng,flag);
	DONE_SUCCESSFULLY();
	return wcc;
}

int fwriteTIMEOUT(PCStr(b),int s,int n,FILE *fp)
{	int rc;

	RETURN_ONTIMEOUT(fileno(fp),0);
	rc = Fwrite(b,s,n,fp);
	DONE_SUCCESSFULLY();
	return rc;
}
int fputsTIMEOUT(PCStr(b),FILE *fp)
{	int rcode;

	RETURN_ONTIMEOUT(fileno(fp),0);
	rcode = fputs(b,fp);
	DONE_SUCCESSFULLY();
	return rcode;
}
extern int LIN_TIMEOUT;
int fflushTIMEOUT(FILE *fp)
{	int rcode;

	if( feof(fp) )
		return EOF;

	RETURN_ONTIMEOUTX(fileno(fp),EOF,LIN_TIMEOUT+1);
	rcode = fflush(fp);
	DONE_SUCCESSFULLY();
	return rcode;
}
int fcloseTIMEOUT(FILE *fp)
{	int rcode;

	if( feof(fp) )
		return EOF;

	RETURN_ONTIMEOUTX(fileno(fp),EOF,LIN_TIMEOUT+2);
	rcode = fclose(fp);
	DONE_SUCCESSFULLY();
	return rcode;
}
FILE *fopenTO(PCStr(path),PCStr(mode),int timeout){
	FILE *fp;

	RETURN_ONTIMEOUTX(-1,NULL,timeout);
	fp = fopen(path,mode);
	DONE_SUCCESSFULLY();
	return fp;
}

static int readSERNO;
int readTimeoutBlocked(int fd,PVStr(buf),int siz,int timeout)
{	int omask,nmask;
	int rcc;
	int serno;

	alertVStr(buf,siz);
	rcc = -1;
	omask = sigblock(sigmask(SIGCHLD));
	serno = ++readSERNO;
	if( 0 < PollIn(fd,timeout) ){
		rcc = read(fd,(char*)buf,siz);
		if( rcc != siz ){
		sv1log("##ERROR: readTimeoutB insufficient read %d/%d (%d)%X\n",
			rcc,siz,errno,sigblock(0));
		}
	}
	if( serno != readSERNO ){
		sv1log("##ERROR: readTimeoutB broken %d/%d (%d)%X\n",
			serno,readSERNO,errno,sigblock(0));
		sleep(10);
	}
	nmask = sigsetmask(omask);
	return rcc;
}
int readTimeout(int fd,PVStr(b),int z,int tout)
{	int nready,rc;

	alertVStr(b,z);
	if( 0 < (nready = xPollIn(fd,tout*1000)) )
		return read(fd,(char*)b,z);
	else{
		sv1log("read(%d) -- ready=%d IO_TIMEOUT(%d)\n",z,nready,tout);
		return 0;
	}
}
int readTIMEOUT(int fd,PVStr(b),int z)
{
	alertVStr(b,z);
	return readTimeout(fd,BVStr(b),z,IO_TIMEOUT);
}
int recvTIMEOUT(int fd,PVStr(b),int z,int f)
{	int rc;

	alertVStr(b,z);
	if( 0 < xPollIn(fd,IO_TIMEOUT*1000) )
		rc = recv(fd,(char*)b,z,f);
	else	rc = 0;
	return rc;
}
static int recvPeekTIMEOUT1(int fd,PVStr(b),int z)
{	int rc;

	alertVStr(b,z);
	RETURN_ONTIMEOUTX(fd,-1,10);
	rc = RecvPeek(fd,(char*)b,z);
	DONE_SUCCESSFULLY();
	return rc;
}
int recvPeekTIMEOUT(int fd,PVStr(b),int z)
{	int rc;

	alertVStr(b,z);
	if( 0 < xPollIn(fd,IO_TIMEOUT*1000) ){
		rc = recvPeekTIMEOUT1(fd,BVStr(b),z);
		if( rc <= 0 )
			sv1log("#### recvPeek: failed: %d\n",rc);
	}else{
		sv1log("recvPeekTIMEOUT: %d\n",IO_TIMEOUT);
		rc = 0;
	}
	return rc;
}
int fgetBuffered(PVStr(b),int n,FILE *fp);
int freadTIMEOUT(PVStr(b),int s,int n,FILE *fp)
{	int rn;

	alertVStr(b,s*n);
	if( feof(fp) ){
		sv1log("-- Tried freadTIMEOUT() for EOF file.\n");
		return 0;
	}
	if( 0 < READYCC(fp) ){
		if( s == 1 ){
			rn = fgetBuffered(BVStr(b),n,fp);
			if( rn == n )
				return rn;
			return rn + freadTIMEOUT(QVStr(b+rn,b),s,n-rn,fp);
		}
	}
	if( 0 < READYCC(fp) || 0 < xPollIn(fileno(fp),IO_TIMEOUT*1000) )
		return fread((char*)b,s,n,fp);
	else{
		sv1log("fread(%d*%d) -- IO_TIMEOUT(%d)\n",s,n,IO_TIMEOUT);
		return 0;
	}
}
void discardBuffered(FILE *fp)
{
	while( READYCC(fp) ){
		getc(fp);
	}
}
int fgetsBuffered(PVStr(b),int n,FILE *fp)
{	int mcc,cc,ch;

	alertVStr(b,n);
	n--;
	for( cc = 0; cc < n; cc++ ){
		if( READYCC(fp) <= 0 )
			break;
		ch = getc(fp);
		setVStrElem(b,cc,ch); /**/
		if( ch == 0 )
			break;
		if( ch == EOF )
			break;
		if( ch == '\n' ){
			cc++;
			break;
		}
	}
	setVStrEnd(b,cc); /**/
	return cc;
}
int fgetBuffered(PVStr(b),int n,FILE *fp)
{	int mcc,cc,ch;

	alertVStr(b,n);
	for( cc = 0; cc < n; cc++ ){
		if( READYCC(fp) <= 0 )
			break;
		ch = getc(fp);
		setVStrElem(b,cc,ch);
		if( ch == EOF )
			break;
	}
	if( cc < n )
		setVStrEnd(b,cc);
	return cc;
}
int file_copyBuffered(FILE *in,FILE *out)
{	int cc,ch;

	for( cc = 0; 0 < READYCC(in) && (ch = getc(in)) != EOF; cc++ ){
		if( putc(ch,out) == EOF )
			break;
	}
	return cc;
}

char *fgets0(PVStr(b),int n,FILE *fp)
{	int ch,li;

	alertVStr(b,n);
	for( li = 0; li < n - 1; li++ ){
		ch = getc(fp);
		setVStrElem(b,li,ch); /**/
		if( ch == EOF || ch == '\n' )
			break;
		if( ch == 0 ){
			ungetc(ch,fp);
			break;
		}
	}
	setVStrEnd(b,li); /**/
	if( li == 0 )
		return NULL;
	else	return (char*)b;
}

static char *fgetsIfReady(PVStr(b),int s,FILE *fp,int *rc)
{	int bx,ch;

	alertVStr(b,s);
	bx = 0;
	while( bx < s-1 && 0 < READYCC(fp) ){
		ch = getc(fp);
		if( ch == EOF )
			break;
		setVStrElemInc(b,bx,ch); /**/
		if( ch == '\n' )
			break;
	}
	setVStrEnd(b,bx); /**/
	if( rc ) *rc = bx;

	if( bx == 0 && feof(fp) )
		return NULL;
	else	return (char*)b;
}

char *fgetsTIMEOUT(xPVStr(b),int s,FILE *fp)
{	const char *b0;
	const char *rs;
	int rc;

	alertVStr(b,s);
	setVStrEnd(b,0);
	if( feof(fp) ){
		sv1log("-- Tried fgetsTIMEOUT() for EOF file.\n");
		return NULL;
	}

	b0 = b;
	rc = 0;
	if( 0 < READYCC(fp) ){
		rs = fgetsIfReady(BVStr(b),s,fp,&rc);
		if( 0 < rc && b[rc-1] == '\n' || feof(fp) || rc == s-1 )
			return (char*)rs;
		b += rc;
		s -= rc;
	}
	if( 0 < xPollIn(fileno(fp),IO_TIMEOUT*1000) )
		rs = Xfgets(BVStr(b),s,fp);
	else{
		CStr(wh,128);
		VStrId(AVStr(wh),BVStr(b));
		sv1log("fgets(%d) -- IO_TIMEOUT(%d) %s\n",s,IO_TIMEOUT,wh);
		/*
		sv1log("fgets(%d) -- IO_TIMEOUT(%d)\n",s,IO_TIMEOUT);
		*/
		rs = NULL;
	}
	if( rs == NULL && rc == 0 )
		return NULL;
	else	return (char*)b0;
}
char *fgetsTimeout(PVStr(b),int s,FILE *fp,int tout)
{	int stout;
	const char *rs;

	alertVStr(b,s);
	if( tout ){
		stout = IO_TIMEOUT;
		IO_TIMEOUT = tout;
		rs = fgetsTIMEOUT(BVStr(b),s,fp);
		IO_TIMEOUT = stout;
	}else	rs = Xfgets(BVStr(b),s,fp);
	return (char*)rs;
}

int fgetcTIMEOUT(FILE *fp)
{
	if( 0 < fPollIn(fp,IO_TIMEOUT*1000) )
		return getc(fp);
	else	return EOF;
}

char *fgetsTO(PVStr(b),int z,FILE *f,int t1,int t2){
	refQStr(bp,b);
	int ch;

	if( fPollIn(f,t1) <= 0 )
		return NULL;
	for(;;){
		ch = getc(f);
		if( ch == EOF )
			break;
		setVStrPtrInc(bp,ch);
		if( ch == '\n' )
			break;
		if( fPollIn(f,t2) <= 0 )
			break;
	}
	setVStrEnd(bp,0);
	if( b < bp ) 
		return (char*)b;
	else	return NULL;
}

char *fgetsByBlock(PVStr(line),int size,FILE *fs,int niced,int ltimeout,int byline,int fromcache,int remlen,int *lengp,int *isbinp)
{	int cc,ch,bc;
	int timeout;
	int stimeout;
	const char *rcode = line;
	int insize;

	alertVStr(line,size);
	if( !fromcache ){
		stimeout = 30;
		timeout = ltimeout;
	}
	if( niced )
		insize = size / 2;
	else	insize = 1024;

	size--;
	bc = 0;

	for( cc = 0; cc < size; ){
		if( !fromcache && READYCC(fs)<1 && fPollIn(fs,timeout)<=0 ){
			if( cc == 0 ){
				syslog_ERROR("fgetsBB: TIMEOUT %dms\n",timeout);
				rcode = NULL;
			}
			break;
		}
		if( (ch = getc(fs)) == EOF ){
			if( cc == 0 )
				rcode = NULL;
			break;
		}

		setVStrElemInc(line,cc,ch); /**/
		if( ch == '\n' && (byline || insize < cc) )
			break;
		if( ch == 0 )
			bc++;

		if( !byline ){
			if( ch == '\n' || ch == '\r' )
				timeout = stimeout;
			else
			if( remlen <= cc ){
				if( timeout != stimeout )
				syslog_DEBUG("fgetsBB: %d/%d TIMEOUT=%d->%d\n",
					cc,remlen,timeout,stimeout);
				timeout = stimeout;
			}
			else	timeout = ltimeout;
		}
	}

	setVStrEnd(line,cc); /**/
	*lengp = cc;
	*isbinp = bc;
	return (char*)rcode;
}

char *fgetsByLine(PVStr(line),int lsiz,FILE *in,int timeout,int *rccp,int *isbinp)
{
	alertVStr(line,lsiz);
	if( fPollIn(in,0) <= 0 )
		return NULL;
	return fgetsByBlock(BVStr(line),lsiz,in,0,timeout,1,0,lsiz,rccp,isbinp);
}

int MAX_NICE_VALUE = 10;
int MAX_NICE_SLEEP = 10;
int MIN_NICE_SLEEP =  1;
int MIN_SOCK_BUFSIZE = 1024;
int MAX_SOCK_BUFSIZE = 1024*64;
/*
 * try to do I/O every MAX_NICE_SLEEP seconds
 * to reduce the cost for process switch
 * without decreasing the speed of relay
 */
int doNice(PCStr(what),int ifd,FILE *ifp,int ofd,FILE *ofp,int niced,int bytes,int count,double since)
{	double now,elapsed;
	double sec1,sec;
	int bsize,isize,osize;

	now = Time();
	elapsed = now - since;

	if( 10 <= (now - since) )
	if( niced < MAX_NICE_VALUE && (0x010000 << niced) < bytes ){
		niced += 1;
		nice(1);

		bsize = ((int)((bytes/elapsed) * MAX_NICE_SLEEP*2)) & ~0x3FF;
		if( bsize == 0 )
			bsize = MIN_SOCK_BUFSIZE;
		if( MAX_SOCK_BUFSIZE < bsize )
			bsize = MAX_SOCK_BUFSIZE;

		if( 0 <= ifd ) expsockbuf(ifd,bsize,0);
		if( 0 <= ofd ) expsockbuf(ofd,0,bsize);

		isize = osize = 0;
		if( 0 <= ifd )
			getsockbuf(ifd,&isize,&osize);
		daemonlog("E","NICE-%d %dK/%d/%4.2fs %d/p %d/s buf=%d(%d)\n",
			niced,bytes/1024,count,elapsed,bytes/count,
			(int)(bytes/elapsed),isize,bsize);
	}
	if( 0 < niced )
	if( ifp == NULL || READYCC(ifp) <= 0 )
	if( 0 <= ifd && !readyAlways(ifd) && PollIn(ifd,1) == 0 )
	if( getsockbuf(ifd,&isize,&osize) == 0 ){
		sec = sec1 = (isize/4) / (bytes/elapsed);
		if( MIN_NICE_SLEEP < sec1 ){
			if( MAX_NICE_SLEEP < sec ) sec = MAX_NICE_SLEEP;
		daemonlog("E","NICE-%d %dK/%d/%4.2fs %d/p %d/s: %4.2fs/%4.2fs\n",
				niced,bytes/1024,count,elapsed,bytes/count,
				(int)(bytes/elapsed),sec,sec1);

			if( ofp != NULL )
				fflush(ofp);
			msleep((int)(sec*1000));
		}
	}
	return niced;
}

int connectToX(PCStr(addr),int port,int delayms,int infd)
{	int rcode,sock,nready,fdv[2],qev[2],rev[2];
	double Start1;
	CStr(stat,64);

	Start1 = Time();
	nready = rev[0] = rev[2] = 0;

	sock = Socket1("SMTP-CB", -1,NULL,NULL,NULL, VStrNULL,0, NULL,0, 0,NULL,0);
	if( sock < 0 ){
		rcode = -1;
		goto EXIT;
	}
	if( connectTimeout(sock,addr,port,1) == 0 ){
		rcode = 0;
		goto EXIT;
	}

	Start1 = Time();
	fdv[0] = infd; qev[0] = PS_IN;
	fdv[1] = sock; qev[1] = PS_OUT|PS_PRI;
	nready = PollInsOuts(delayms,2,fdv,qev,rev);

	if( nready <= 0 ){
		rcode = -2;
		goto EXIT;
	}
	if( (rev[1] & PS_OUT) == 0 ){
		rcode = -3;
		goto EXIT;
	}
	if( IsAlive(sock) ){
		rcode = 0;
		goto EXIT;
	}
	rcode = -4;

EXIT:
	if( rcode != 0 ){
		if( 0 <= sock ){
			close(sock);
			sock = -1;
		}
	}

	sprintf(stat,"%d[%d %s %s]",
		sock,
		sock_isconnected(sock),
		IsConnected(sock,NULL)?"Connected":"-",
		IsAlive(sock)?"Alive":"-");

	sv1log("#CToX%d %s %s:%d %d[%o][%o] %s %dms\n",
		rcode,rcode==0?"OK":"ERR", addr,port,
		nready,rev[0],rev[1],
		stat,(int)(1000*(Time()-Start1)));

	return sock;
}

