/*////////////////////////////////////////////////////////////////////////
Copyright (c) 2005 National Institute of Advanced Industrial Science and Technology (AIST)

Permission to use this material for evaluation, copy this material for
your own use, and distribute the copies via publically accessible on-line
media, without fee, is hereby granted provided that the above copyright
notice and this permission notice appear in all copies.
AIST MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	sftp.c (ftp client to sftp server gateway)
Author:		Yutaka Sato <ysato@delegate.org>
Description:
	PROBLEM
		uploading huge file (STOR)
		 ... getting into local tmpfile is bad
History:
	050728	created
//////////////////////////////////////////////////////////////////////#*/
#include "vsocket.h"
#include "dglib.h"
#include "fpoll.h"
#include "proc.h"

#include "log.h"
#define DEBUG	lGATEWAY()?0:sv1log
#define Error	sv1log

int CC_connect(PCStr(proto),PCStr(host),int port,PCStr(user));
int CC_accept(PCStr(proto),PCStr(host),int port,PCStr(user),int fromS);
int Forkpty(int *pty,char *name);
int Mkfifo(PCStr(path),int mode);
char *fgetsTO(PVStr(b),int z,FILE *f,int t1,int t2);
FILE *fopenTO(PCStr(path),PCStr(mode),int timeout);
int file_copy(FILE *src,FILE *dst,FILE *cache,int bytes,int *binary);

static int DATA_MAX = 10*1024*1024;
static char *sftp_com = "sftp";
static int pid;

static void putresp(FILE *tc,int code,PCStr(stat),...){
	CStr(resp,1024);
	refQStr(rp,resp);
	VARGS(8,stat);

	sprintf(rp,"%d ",code); rp += strlen(rp);
	sprintf(rp,stat,VA8);
	fprintf(tc,"%s\r\n",resp);
	fflush(tc);
	DEBUG("--SFTPGW >> %s\n",resp);
}
static void putsftp(FILE *ts,PCStr(fmt),...){
	CStr(com,1024);
	VARGS(8,fmt);

	sprintf(com,fmt,VA8);
	fprintf(ts,"%s\r\n",com);
	fflush(ts);
	DEBUG("--SFTP << %s\r\n",com);
}

static void relay_resp(FILE *fs,int dsock,int timeout,PVStr(resp),PCStr(com),int dolog){
	CStr(res,1024);
	FILE *fp;
	refQStr(rp,resp);
	refQStr(dp,res);
	int tout0;
	int tout;

	if( 0 < timeout )
		tout0 = tout = timeout;
	else	tout0 = tout = 3*1000;

	if( resp )
		setVStrEnd(resp,0);
	if( fPollIn(fs,tout) <= 0 ){
		DEBUG("--SFTP >> response timeout [%s]%d\n",com,tout);
		return;
	}
	if( 0 <= dsock )
		fp = fdopen(dsock,"w");
	else	fp = NULL;

	for(;;){
		if( 0 < fgetsBuffered(AVStr(res),sizeof(res),fs) ){
		}else
		if( fgetsTO(AVStr(res),sizeof(res),fs,tout,1) == NULL ){
			break;
		}
		if( dolog ){
			if( strchr(res,'\n') )
				DEBUG("--SFTP >>(%s) %s",com,res);
			else	DEBUG("--SFTP >>(%s/NO-NL)[%s]\n",com,res);
		}
		if( strtailstr(res,"sftp> ") ){
			if( fPollIn(fs,10) <= 0 )
				break;
		}
		if( dp = strpbrk(res,"\r\n") )
			strcpy(dp,"\r\n");
		if( fp != NULL )
			fputs(res,fp);
		if( rp ){
			strcpy(rp,res);
			rp += strlen(rp);
		}
		if( strtailstr(res,":")
		 || strtailstr(res,": ")
		 || strtailstr(res,"? ")
		){
			tout = 1;
		}else
		if( *res == '\r' || *res == '\n' )
			tout = 100;
		else	tout = tout0;
	}
	if( fp != NULL )
		fflush(fp);
}

static void closeAll(){
	int fi;
	int rcode;
	for( fi = 3; fi < 256; fi++ ){
		rcode = close(fi);
		/*
		if( rcode == 0 ) DEBUG("-- closed %d\n",fi);
		*/
	}
}
static int forkSftp(PCStr(host),int port,PCStr(user),PCStr(pass),int tf[2]){
	CStr(com,1024);
	CStr(resp,0x10000);
	int pty;
	char name[128];
	FILE *fs;
	FILE *ts;
	int to[2];
	int from[2];
	int rcode;

	sprintf(com,"%s -oPort=%d %s@%s",sftp_com,port,user,host);
	DEBUG("--SFTP: %s\n",com);

	pid = Forkpty(&pty,name);
	if( 0 < pid ){
		DEBUG("--SFTP: pty master %d %s\n",pty,name);
		tf[0] = pty;
		tf[1] = pty;

		msleep(100);
		fs = fdopen(pty,"r");
		relay_resp(fs,-1,15*1000,AVStr(resp),"FORK-1",1);

		if( strstr(resp,"Are you sure you want to continue") ){
			sprintf(com,"yes\n");
			write(pty,com,strlen(com));
			DEBUG("--SFTP: answered yes for...\n%s\n",resp);
			relay_resp(fs,-1,15*1000,AVStr(resp),"FORK-2",1);
		}

		sprintf(com,"%s\r\n",pass);
		write(pty,com,strlen(com));
		DEBUG("--SFTP: sent the password...\n");
		relay_resp(fs,-1,15*1000,AVStr(resp),"FORK-3",1);

		sprintf(com,"progress\r\n");
		write(pty,com,strlen(com));
		relay_resp(fs,-1,5*1000,AVStr(resp),"FORK-4",1);

		if( strtailstr(resp,"Password:")
		 || strtailstr(resp,"password: ")
		 || strstr(resp,"Enter passphrase for key")
		){
			DEBUG("--SFTP %s\n",resp);
			return -1;
		}
		return 0;
	}
	if( pid == 0 ){
		closeAll();
		system("stty raw");
		rcode = system(com);
		printf("ssh exit(%d)\n",rcode);
		_exit(0);
	}

	pipe(to);
	pipe(from);
	if( Fork("Sftp") == 0 ){
		close(to[1]);
		dup2(to[0],0);
		close(from[0]);
		dup2(from[1],1);
		closeAll();
		rcode = system(com);
		printf("ssh exit(%d)\n",rcode);
		_exit(0);
	}
	close(to[0]);
	close(from[1]);
	tf[0] = from[0];
	tf[1] = to[1];
	fs = fdopen(from[0],"r");
	relay_resp(fs,-1,20*1000,AVStr(resp),"FORK-1",1);
	return 0;
}

static int relay_fifo(FILE *fs,FILE *tc,PCStr(fifo),int dsock,PCStr(com)){
	FILE *sfp;
	FILE *dfp;
	int fv[2];
	int rv[2];
	int nready;

	sfp = fopenTO(fifo,"r",1);
	DEBUG("--SFTP relay [%s] %X\n",fifo,sfp);
	if( sfp == NULL ){
		relay_resp(fs,-1,1,VStrNULL,"RESP-A",1);
		return -1;
	}

	fv[0] = fileno(fs);
	fv[1] = fileno(sfp);

	for(;;){
		nready = PollIns(1000,2,fv,rv);
		DEBUG("--SFTP poll [%s] %d[%d %d]\n",fifo,nready,rv[0],rv[1]);
		if( nready <= 0 ){
			DEBUG("--SFTP NO RESP-B\n");
			break;
		}else{
			if( 0 < rv[0] ){
				relay_resp(fs,-1,1,VStrNULL,com,1);
			}
			if( 0 < rv[1] ){
				dfp = fdopen(dsock,"w");
				if( dfp ){
					int isbin;
					int bytes;
					DEBUG("--SFTP relay [%s]...\n",fifo);
					putresp(tc,150,"Ok");
					bytes=file_copy(sfp,dfp,NULL,0,&isbin);
					fcloseFILE(dfp);
					DEBUG("--SFTP bin=%d,%d downloaded\n",
						isbin,bytes);
				}
				break;
			}
		}
	}
	fclose(sfp);
	return 0;
}

unsigned int trand1(unsigned int max);
static FILE *getfile(PVStr(path),PCStr(mode),int fifo){
	CStr(npath,1024);
	int max = 64;
	FILE *fp;
	int fi;
	int fx;

	sprintf(path,"/tmp/sftpgw.%d.%d",getpid(),time(NULL));
	if( fifo ){
		if( Mkfifo(path,0600) != 0 ){
			DEBUG("--SFTP cant create FIFO: %s\n",path);
			return 0;
		}
		fp = NULL;
	}else{
		if( (fp = fopen(path,mode)) == 0 ){
			DEBUG("--SFTP cant create FILE: %s\n",path);
			return 0;
		}
	}
	fx = trand1(max);
	for( fi = 0; fi < max; fi++ ){
		sprintf(npath,"/tmp/sftpgw.%02x",fx);
		if( rename(path,npath) == 0 ){
			DEBUG("--SFTP TEMP FILE: %s\n",npath);
			strcpy(path,npath);
			return fp;
		}
	}
	DEBUG("--SFTP ERROR: can't rename %s\n",path);
	return fp;
}

static void SftpGW(PCStr(host),int port,int gw){
	int tofrom_sftp[2];
	FILE *fc;
	FILE *tc;
	FILE *ts;
	FILE *fs;
	int rcc;
	CStr(req,1024);
	CStr(com,1024);
	CStr(arg,1024);
	CStr(user,128);
	CStr(passMD5,128);
	CStr(tmp,1024);
	CStr(logindir,1024);
	CStr(rnfr,512);
	CStr(res,1024);
	CStr(xcom,1024);
	const char *dp;
	int vsock = -1;
	int dsock = -1;
	CStr(fifo,1024);
	CStr(sresp,0x10000);
	CStr(resp1,1024);
	int Ok;
	int Bad;
	int xpid;
	int nready;
	int nfc;
	FILE *fpv[2];
	int rdv[2];
	int ncc = 0;

	DEBUG("--SFTPGW start\n");
	fc = fdopen(gw,"r");
	tc = fdopen(gw,"w");
	putresp(tc,220,"SFTP/FTP gateway ready.");
	fflush(tc);
	ts = NULL;
	fs = NULL;

	logindir[0] = 0;
	fpv[0] = fc;
	nfc = 1;
	for(;;){
		nready = fPollIns(0,nfc,fpv,rdv);
		if( nready == 0 && errno == EINTR ){
			DEBUG("--SFTPGW ignored EINTR\n");
			sleep(1);
			continue;
		}
		if( nready <= 0 ){
			break;
		}
		if( 1 < nfc && 0 < rdv[1] ){
			DEBUG("--SFTPGW EOF from server?\n");
			relay_resp(fs,-1,0,AVStr(sresp),"",1);
			break;
		}

		fflush(tc);
		if( fgets(req,sizeof(req),fc) == 0 )
			break;
		dp = wordScan(req,com);
		lineScan(dp,arg);
		if( strcaseeq(com,"PASS") )
			DEBUG("--SFTPGW << [%s][****]\n",com);
		else	DEBUG("--SFTPGW << [%s][%s]\n",com,arg);

		if( strcaseeq(com,"QUIT") ){
			putresp(tc,221,"Ok Bye.");
			fclose(fc);
			fclose(tc);
			fc = 0;
			tc = 0;
			gw = CC_accept("sftp",host,port,user,fileno(fs));
			if( 0 <= gw ){
				fc = fdopen(gw,"r");
				tc = fdopen(gw,"w");
				ncc++;
				DEBUG("--SFTPGW SFTPCC restart #%d\n",ncc);
				putresp(tc,220,"Ok (reusing)");
				continue;
			}
			DEBUG("--SFTPGW DONE\n");
			break;
		}else
		if( strcaseeq(com,"NOOP") ){
			putresp(tc,200,"Ok");
		}else
		if( strcaseeq(com,"USER") ){
			strcpy(user,arg);
			putresp(tc,331,"Send password or passphrase for '%s'",
				arg);
		}else
		if( strcaseeq(com,"PASS") ){
			if( ts != NULL ){
				toMD5(arg,tmp);
				DEBUG("--SFTP reusing user[%s]pass[%s][%s]\n",
					user,passMD5,tmp);
				if( !streq(tmp,passMD5) ){
					putresp(tc,530,"No");
					continue;
				}else	putresp(tc,230,"Ok");
				putsftp(ts,"cd %s",logindir);
				relay_resp(fs,-1,0,AVStr(sresp),"restart",1);
				continue;
			}
			if( forkSftp(host,port,user,arg,tofrom_sftp) != 0 ){
				DEBUG("--SFTP login failed\n");
				putresp(tc,530,"No (Login failed)");
				break;
			}
			toMD5(arg,passMD5);
			fs = fdopen(tofrom_sftp[0],"r");
			ts = fdopen(tofrom_sftp[1],"w");
			fpv[1] = fs;
			nfc = 2;
			putresp(tc,230,"Ok");

			putsftp(ts,"pwd");
			relay_resp(fs,-1,0,AVStr(sresp),com,1);
			if( strneq(sresp,"Remote working directory: ",26) ){
				CStr(path,1024);
				lineScan(sresp+26,path);
				strcpy(logindir,path);
			}
		}else
		if( strcaseeq(com,"PWD") ){
			putsftp(ts,"pwd");
			relay_resp(fs,-1,0,AVStr(sresp),com,1);
			if( strneq(sresp,"Remote working directory: ",26) ){
				CStr(path,1024);
				lineScan(sresp+26,path);
				putresp(tc,257,"\"%s\"",path);
			}else{
				putresp(tc,257,"\"/\"");
			}
		}else
		if( strcaseeq(com,"CWD") ){
			putsftp(ts,"cd %s",arg);
			relay_resp(fs,-1,0,AVStr(sresp),com,1);
			if( strneq(sresp,"Couldn't",8)
			 || strneq(sresp,"Can't",4) ){
				putresp(tc,550,"No");
			}else	putresp(tc,250,"Ok");
		}else
		if( strcaseeq(com,"TYPE") ){
			putresp(tc,200,"Ok");
		}else
		if( strcaseeq(com,"PASV") ){
			VSAddr ba;
			int balen;
			CStr(mport,128);
			vsock = server_open("SftpGW",VStrNULL,0,1);
			if( 0 <= vsock ){
				balen = sizeof(ba);
				getsockname(vsock,(SAP)&ba,&balen);
				VSA_prftp(&ba,AVStr(mport));
			putresp(tc,227,"Enterning Passive Mode (%s)",mport);
			}else{
				putresp(tc,500,"No");
			}
		}else
		if( strcaseeq(com,"STOR") ){
			if( vsock < 0 ){
				putresp(tc,500,"No");
				continue;
			}
			dsock = ACCEPT(vsock,1,-1,10);
			if( 0 <= dsock ){
				CStr(tmpf,1024);
				FILE *dfp;
				FILE *tmp;
				int bytes;
				int isbin;

				dfp = fdopen(dsock,"r");
				if( dfp == NULL ){
					putresp(tc,500,"No-1");
					continue;
				}
				tmp = getfile(AVStr(tmpf),"w",0);
				if( tmp == NULL ){
					putresp(tc,500,"No-2");
					fclose(dfp);
					continue;
				}
				putresp(tc,150,"Ok");
				bytes = file_copy(dfp,tmp,NULL,DATA_MAX,&isbin);
				DEBUG("--SFTPGW bin=%d,%d bytes upload\n",
					isbin,bytes);
				fclose(tmp);
				fclose(dfp);

				putsftp(ts,"put %s %s",tmpf,arg);
				relay_resp(fs,-1,15*1000,AVStr(sresp),com,1);
				if( strncmp(sresp,"Could",5) == 0 )
					putresp(tc,550,"No");
				else	putresp(tc,226,"Ok");
				unlink(tmpf);
				dsock = -1;
			}else{
				putresp(tc,500,"accept error");
			}
			vsock = -1;
		}else
		if( strcaseeq(com,"RETR") ){
			if( vsock < 0 ){
				putresp(tc,500,"No");
				continue;
			}
			dsock = ACCEPT(vsock,1,-1,10);
			if( arg[0]==0 ||streq(arg,".") ||strtailstr(arg,"/") ){
				putresp(tc,550,"RETR for dir [%s]",arg);
				if( 0 <= dsock ){
					close(dsock);
					dsock = -1;
				}
			}else
			if( 0 <= dsock ){
				getfile(AVStr(fifo),NULL,1);
				putsftp(ts,"get %s %s",arg,fifo);
				if( relay_fifo(fs,tc,fifo,dsock,com)==0 )
					putresp(tc,226,"Ok");
				else	putresp(tc,550,"No");
				relay_resp(fs,-1,1,AVStr(sresp),com,1);
				close(dsock);
				dsock = -1;
				unlink(fifo);
			}else{
				putresp(tc,500,"accept error");
			}
			vsock = -1;
		}else
		if( strcaseeq(com,"NLST") || strcaseeq(com,"LIST") ){
			if( vsock < 0 ){
				putresp(tc,500,"No");
				continue;
			}
			dsock = ACCEPT(vsock,1,-1,10);
			if( 0 <= dsock ){
				putresp(tc,150,"Ok");
				if( arg[0] == '-' ){
					char *op;
					for( op = arg+1; *op; ){
						if( strchr(" \t\r\n",*op) )
							break;
						if( strchr("L",*op) ){
							ovstrcpy(op,op+1);
						}else	op++;
					}
				}
				if( streq(com,"NLST") || strstr(arg,"-l") )
					putsftp(ts,"ls %s",arg);
				else	putsftp(ts,"ls -l %s",arg);
				relay_resp(fs,dsock,0,VStrNULL,com,0);
				close(dsock);
				dsock = -1;
				putresp(tc,226,"Ok");
			}else{
				putresp(tc,500,"accept error");
			}
			vsock = -1;
		}else
		if( strcaseeq(com,"RNFR") ){
			strcpy(rnfr,arg);
			putresp(tc,350,"Ok");
		}else
		if( strcaseeq(com,"RNTO") ){
			putsftp(ts,"rename %s %s",rnfr,arg);
			relay_resp(fs,-1,0,AVStr(sresp),com,1);
			if( dp = strstr(sresp,"Couldn't") ){
				lineScan(dp,resp1);
				putresp(tc,550,"No %s",resp1);
			}else	putresp(tc,250,"Ok");
		}else
		if( strcaseeq(com,"MKD") || strcaseeq(com,"RMD")
		 || strcaseeq(com,"DELE")
		){ 
			Bad = 550;
			if( strcaseeq(com,"MKD") ){
				Ok = 257;
				putsftp(ts,"mkdir %s",arg);
			}else
			if( strcaseeq(com,"RMD") ){
				Ok = 250;
				putsftp(ts,"rmdir %s",arg);
			}else{
				Ok = 250;
				putsftp(ts,"rm %s",arg);
			}
			relay_resp(fs,-1,0,AVStr(sresp),com,1);
			if( dp = strstr(sresp,"Couldn't") ){
				lineScan(dp,resp1);
				putresp(tc,Bad,"No %s",resp1);
			}else	putresp(tc,Ok,"Ok");
		}else
		{
			putresp(tc,502,"No not supported");
		}
	}
	if( fs != NULL && !feof(fs) ){
		putsftp(ts,"quit");
		relay_resp(fs,-1,0,AVStr(sresp),com,1);
	}
	xpid = NoHangWait();
	DEBUG("--SFTPGW finished, pid=%d\n",xpid);
}
int connectToSftp(const char *host,int port,const char *user,int fdc,int fdv[]){
	int socks[2];

	socks[1] = CC_connect("sftp",host,port,user);
	if( 0 <= socks[1] ){
		DEBUG("---- SFTPCC HIT[%d] %s@%s:%d\n",socks[1],user,host,port);
		return socks[1];
	}
	DEBUG("---- SFTPCC MISS %s@%s:%d\n",user,host,port);
	Socketpair(socks);
	if( Fork("SftpGW") == 0 ){
		int fi;
		for( fi = 0; fi < fdc; fi++ )
			close(fdv[fi]);
		close(socks[1]);
		SftpGW(host,port,socks[0]);
		_exit(0);
	}
	close(socks[0]);
	return socks[1];
}
